@@ -40,8 +40,11 @@ mod de;
 mod error;
 mod ser;
 
+#[allow(deprecated)]
 pub use crate::de::depythonize;
+pub use crate::de::{depythonize_bound, Depythonizer};
 pub use crate::error::{PythonizeError, Result};
 pub use crate::ser::{
-    pythonize, pythonize_custom, PythonizeDictType, PythonizeListType, PythonizeTypes,
+    pythonize, pythonize_custom, PythonizeDefault, PythonizeDictType, PythonizeListType,
+    PythonizeTypes, Pythonizer,
 };
