@@ -1,7 +1,7 @@
 use std::marker::PhantomData;
 
-use pyo3::types::{PyDict, PyList, PyMapping, PySequence, PyTuple};
-use pyo3::{IntoPy, PyObject, PyResult, Python, ToPyObject};
+use pyo3::types::{PyAnyMethods, PyDict, PyList, PyMapping, PySequence, PyTuple};
+use pyo3::{Bound, IntoPy, PyObject, PyResult, Python, ToPyObject};
 use serde::{ser, Serialize};
 
 use crate::error::{PythonizeError, Result};
@@ -9,7 +9,7 @@ use crate::error::{PythonizeError, Result};
 /// Trait for types which can represent a Python mapping
 pub trait PythonizeDictType {
     /// Constructor
-    fn create_mapping(py: Python) -> PyResult<&PyMapping>;
+    fn create_mapping(py: Python) -> PyResult<Bound<PyMapping>>;
 }
 
 /// Trait for types which can represent a Python sequence
@@ -18,7 +18,7 @@ pub trait PythonizeListType: Sized {
     fn create_sequence<T, U>(
         py: Python,
         elements: impl IntoIterator<Item = T, IntoIter = U>,
-    ) -> PyResult<&PySequence>
+    ) -> PyResult<Bound<PySequence>>
     where
         T: ToPyObject,
         U: ExactSizeIterator<Item = T>;
@@ -33,8 +33,8 @@ pub trait PythonizeTypes {
 }
 
 impl PythonizeDictType for PyDict {
-    fn create_mapping(py: Python) -> PyResult<&PyMapping> {
-        Ok(PyDict::new(py).as_mapping())
+    fn create_mapping(py: Python) -> PyResult<Bound<PyMapping>> {
+        Ok(PyDict::new_bound(py).into_any().downcast_into().unwrap())
     }
 }
 
@@ -42,16 +42,19 @@ impl PythonizeListType for PyList {
     fn create_sequence<T, U>(
         py: Python,
         elements: impl IntoIterator<Item = T, IntoIter = U>,
-    ) -> PyResult<&PySequence>
+    ) -> PyResult<Bound<PySequence>>
     where
         T: ToPyObject,
         U: ExactSizeIterator<Item = T>,
     {
-        Ok(PyList::new(py, elements).as_sequence())
+        Ok(PyList::new_bound(py, elements)
+            .into_any()
+            .downcast_into()
+            .unwrap())
     }
 }
 
-struct PythonizeDefault;
+pub struct PythonizeDefault;
 
 impl PythonizeTypes for PythonizeDefault {
     type Map = PyDict;
@@ -63,7 +66,7 @@ pub fn pythonize<T>(py: Python, value: &T) -> Result<PyObject>
 where
     T: ?Sized + Serialize,
 {
-    pythonize_custom::<PythonizeDefault, _>(py, value)
+    value.serialize(Pythonizer::new(py))
 }
 
 /// Attempt to convert the given data into a Python object.
@@ -73,18 +76,37 @@ where
     T: ?Sized + Serialize,
     P: PythonizeTypes,
 {
-    value.serialize(Pythonizer::<P> {
-        py,
-        _types: PhantomData,
-    })
+    value.serialize(Pythonizer::custom::<P>(py))
 }
 
+/// A structure that serializes Rust values into Python objects
 #[derive(Clone, Copy)]
 pub struct Pythonizer<'py, P> {
     py: Python<'py>,
     _types: PhantomData<P>,
 }
 
+impl<'py, P> From<Python<'py>> for Pythonizer<'py, P> {
+    fn from(py: Python<'py>) -> Self {
+        Self {
+            py,
+            _types: PhantomData,
+        }
+    }
+}
+
+impl<'py> Pythonizer<'py, PythonizeDefault> {
+    /// Creates a serializer to convert data into a Python object using the default mapping class
+    pub fn new(py: Python<'py>) -> Self {
+        Self::from(py)
+    }
+
+    /// Creates a serializer to convert data into a Python object using a custom mapping class
+    pub fn custom<P>(py: Python<'py>) -> Pythonizer<'py, P> {
+        Pythonizer::from(py)
+    }
+}
+
 #[doc(hidden)]
 pub struct PythonCollectionSerializer<'py, P> {
     items: Vec<PyObject>,
@@ -107,14 +129,14 @@ pub struct PythonStructVariantSerializer<'py, P: PythonizeTypes> {
 #[doc(hidden)]
 pub struct PythonDictSerializer<'py, P: PythonizeTypes> {
     py: Python<'py>,
-    dict: &'py PyMapping,
+    dict: Bound<'py, PyMapping>,
     _types: PhantomData<P>,
 }
 
 #[doc(hidden)]
 pub struct PythonMapSerializer<'py, P: PythonizeTypes> {
     py: Python<'py>,
-    map: &'py PyMapping,
+    map: Bound<'py, PyMapping>,
     key: Option<PyObject>,
     _types: PhantomData<P>,
 }
@@ -231,7 +253,7 @@ impl<'py, P: PythonizeTypes> ser::Serializer for Pythonizer<'py, P> {
     where
         T: ?Sized + Serialize,
     {
-        let d = PyDict::new(self.py);
+        let d = PyDict::new_bound(self.py);
         d.set_item(variant, value.serialize(self)?)?;
         Ok(d.into())
     }
@@ -344,7 +366,7 @@ impl<'py, P: PythonizeTypes> ser::SerializeTuple for PythonCollectionSerializer<
     }
 
     fn end(self) -> Result<PyObject> {
-        Ok(PyTuple::new(self.py, self.items).into())
+        Ok(PyTuple::new_bound(self.py, self.items).into())
     }
 }
 
@@ -376,7 +398,7 @@ impl<'py, P: PythonizeTypes> ser::SerializeTupleVariant for PythonTupleVariantSe
     }
 
     fn end(self) -> Result<PyObject> {
-        let d = PyDict::new(self.inner.py);
+        let d = PyDict::new_bound(self.inner.py);
         d.set_item(self.variant, ser::SerializeTuple::end(self.inner)?)?;
         Ok(d.into())
     }
@@ -445,7 +467,7 @@ impl<'py, P: PythonizeTypes> ser::SerializeStructVariant for PythonStructVariant
     }
 
     fn end(self) -> Result<PyObject> {
-        let d = PyDict::new(self.inner.py);
+        let d = PyDict::new_bound(self.inner.py);
         d.set_item(self.variant, self.inner.dict)?;
         Ok(d.into())
     }
@@ -455,9 +477,10 @@ impl<'py, P: PythonizeTypes> ser::SerializeStructVariant for PythonStructVariant
 mod test {
     use super::pythonize;
     use maplit::hashmap;
+    use pyo3::prelude::*;
+    use pyo3::pybacked::PyBackedStr;
     use pyo3::types::PyDict;
-    use pyo3::{PyResult, Python};
-    use serde::{Deserialize, Serialize};
+    use serde::Serialize;
 
     fn test_ser<T>(src: T, expected: &str)
     where
@@ -466,15 +489,16 @@ mod test {
         Python::with_gil(|py| -> PyResult<()> {
             let obj = pythonize(py, &src)?;
 
-            let locals = PyDict::new(py);
+            let locals = PyDict::new_bound(py);
             locals.set_item("obj", obj)?;
 
-            py.run(
+            py.run_bound(
                 "import json; result = json.dumps(obj, separators=(',', ':'))",
                 None,
-                Some(locals),
+                Some(&locals),
             )?;
-            let result = locals.get_item("result")?.unwrap().extract::<&str>()?;
+            let result = locals.get_item("result")?.unwrap();
+            let result = result.extract::<PyBackedStr>()?;
 
             assert_eq!(result, expected);
             assert_eq!(serde_json::to_string(&src).unwrap(), expected);
@@ -486,7 +510,7 @@ mod test {
 
     #[test]
     fn test_empty_struct() {
-        #[derive(Serialize, Deserialize)]
+        #[derive(Serialize)]
         struct Empty;
 
         test_ser(Empty, "null");
@@ -494,7 +518,7 @@ mod test {
 
     #[test]
     fn test_struct() {
-        #[derive(Serialize, Deserialize)]
+        #[derive(Serialize)]
         struct Struct {
             foo: String,
             bar: usize,
@@ -511,7 +535,7 @@ mod test {
 
     #[test]
     fn test_tuple_struct() {
-        #[derive(Serialize, Deserialize)]
+        #[derive(Serialize)]
         struct TupleStruct(String, usize);
 
         test_ser(TupleStruct("foo".to_string(), 5), r#"["foo",5]"#);
@@ -534,7 +558,7 @@ mod test {
 
     #[test]
     fn test_enum_unit_variant() {
-        #[derive(Serialize, Deserialize)]
+        #[derive(Serialize)]
         enum E {
             Empty,
         }
@@ -544,7 +568,7 @@ mod test {
 
     #[test]
     fn test_enum_tuple_variant() {
-        #[derive(Serialize, Deserialize)]
+        #[derive(Serialize)]
         enum E {
             Tuple(i32, String),
         }
@@ -554,7 +578,7 @@ mod test {
 
     #[test]
     fn test_enum_newtype_variant() {
-        #[derive(Serialize, Deserialize)]
+        #[derive(Serialize)]
         enum E {
             NewType(String),
         }
@@ -564,7 +588,7 @@ mod test {
 
     #[test]
     fn test_enum_struct_variant() {
-        #[derive(Serialize, Deserialize)]
+        #[derive(Serialize)]
         enum E {
             Struct { foo: String, bar: usize },
         }
@@ -577,4 +601,53 @@ mod test {
             r#"{"Struct":{"foo":"foo","bar":5}}"#,
         );
     }
+
+    #[test]
+    fn test_integers() {
+        #[derive(Serialize)]
+        struct Integers {
+            a: i8,
+            b: i16,
+            c: i32,
+            d: i64,
+            e: u8,
+            f: u16,
+            g: u32,
+            h: u64,
+        }
+
+        test_ser(
+            Integers {
+                a: 1,
+                b: 2,
+                c: 3,
+                d: 4,
+                e: 5,
+                f: 6,
+                g: 7,
+                h: 8,
+            },
+            r#"{"a":1,"b":2,"c":3,"d":4,"e":5,"f":6,"g":7,"h":8}"#,
+        )
+    }
+
+    #[test]
+    fn test_bool() {
+        test_ser(true, "true");
+        test_ser(false, "false");
+    }
+
+    #[test]
+    fn test_none() {
+        #[derive(Serialize)]
+        struct S;
+
+        test_ser((), "null");
+        test_ser(S, "null");
+    }
+
+    #[test]
+    fn test_bytes() {
+        test_ser(b"foo", "[102,111,111]");
+    }
 }
